//this file is part of eMule
//Copyright (C)2002 Merkur ( merkur-@users.sourceforge.net / http://www.emule-project.net )
//
//This program is free software; you can redistribute it and/or
//modify it under the terms of the GNU General Public License
//as published by the Free Software Foundation; either
//version 2 of the License, or (at your option) any later version.
//
//This program is distributed in the hope that it will be useful,
//but WITHOUT ANY WARRANTY; without even the implied warranty of
//MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//GNU General Public License for more details.
//
//You should have received a copy of the GNU General Public License
//along with this program; if not, write to the Free Software
//Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
// ListenSocket.cpp : implementation file
//

#include "stdafx.h"
#include "emule.h"
#include "ListenSocket.h"
#include "opcodes.h"
#include "KnownFile.h"
#include "sharedfilelist.h"
#include "uploadqueue.h"
#include "updownclient.h"
#include "clientlist.h"


// CClientReqSocket
CClientReqSocket::CClientReqSocket(CPreferences* in_prefs,CUpDownClient* in_client){
	app_prefs = in_prefs;
	client = in_client;
	if (in_client)
		client->socket = this;
	theApp.listensocket->AddSocket(this);
	ResetTimeOutTimer();
	deletethis = false;
	deltimer = 0;
}


CClientReqSocket::~CClientReqSocket(){
	if (client)
		client->socket = 0;
	client = 0;
	theApp.listensocket->RemoveSocket(this);

	DEBUG_ONLY (theApp.clientlist->Debug_SocketDeleted(this));
}

void CClientReqSocket::ResetTimeOutTimer(){
	timeout_timer = ::GetTickCount();
};

bool CClientReqSocket::CheckTimeOut(){
	if (::GetTickCount() - timeout_timer > CONNECTION_TIMEOUT){
		timeout_timer = ::GetTickCount();
		Disconnect();
		return true;
	}
	return false;
};

void CClientReqSocket::OnClose(int nErrorCode){
	ASSERT (theApp.listensocket->IsValidSocket(this));
	CEMSocket::OnClose(nErrorCode);
	Disconnect();
};

void CClientReqSocket::Disconnect(){
	AsyncSelect(0);
	connected = false;
	if (!client)
		Safe_Delete();
	else
		client->Disconnected();
};

void CClientReqSocket::Delete_Timed(){
// it seems that MFC Sockets call socketfunctions after they are deleted, even if the socket is closed
// and select(0) is set. So we need to wait some time to make sure this doesn't happens
	if (::GetTickCount() - deltimer > 30000)
		delete this;
}

void CClientReqSocket::Safe_Delete(){
	ASSERT (theApp.listensocket->IsValidSocket(this));
	AsyncSelect(0);
	deltimer = ::GetTickCount();
	if (m_hSocket != INVALID_SOCKET)
		ShutDown(2);
	if (client)
		client->socket = 0;
	client = 0;
	connected = false;
	deletethis = true;
}

bool CClientReqSocket::ProcessPacket(char* packet, uint32 size, uint8 opcode){
	try{
		try{
			if (!client && opcode != OP_HELLO)
				throw GetResString(IDS_ERR_NOHELLO);
			switch(opcode){
				case OP_HELLOANSWER:{
					client->ProcessHelloAnswer(packet,size);
					if (client)
						client->ConnectionEstablished();
					break;
				}
				case OP_HELLO:{
					if (!client){
						// create new client to save standart informations
						client = new CUpDownClient(this);
					}
					client->ProcessHelloPacket(packet,size);
					// now we check if we now this client already. if yes this socket will
					// be attached to the known client, the new client will be deleted
					// and the var. "client" will point to the known client.
					// if not we keep our new-constructed client ;)
					if (theApp.clientlist->AttachToAlreadyKnown(&client,this)){
						// update the old client informations
						client->ProcessHelloPacket(packet,size);
					}
					else
						theApp.clientlist->AddClient(client);
					// send a response packet with standart informations
					if (client->GetClientSoft() == SO_EMULE)
						client->SendMuleInfoPacket(false);
					client->SendHelloAnswer();
					if (client)
						client->ConnectionEstablished();

					break;
				}
				case OP_FILEREQUEST:{
					if (size == 16){
						if (!client->GetWaitStartTime())
							client->SetWaitStartTime();
						CKnownFile* reqfile = theApp.sharedfiles->GetFileByID((uchar*)packet);
						if (!reqfile)
							break;
						// if wer are downloading this file, this could be a new source
						if (reqfile->IsPartFile())
							if( theApp.glob_prefs->GetMaxSourcePerFile() > ((CPartFile*)reqfile)->GetSourceCount() ) //<<--
								theApp.downloadqueue->CheckAndAddKnownSource((CPartFile*)reqfile,client);
						
						
						// send filename etc
						memcpy(client->reqfileid,packet,16);
						CMemFile* data = new CMemFile();
						data->Write(reqfile->GetFileHash(),16);
						uint16 namelength = (uint16)strlen(reqfile->GetFileName());
						data->Write(&namelength,2);
						data->Write(reqfile->GetFileName(),namelength);
						Packet* packet = new Packet(data);
						packet->opcode = OP_FILEREQANSWER;
						delete data;
						SendPacket(packet,true);
						//send filestatus
						data = new CMemFile();
						data->Write(reqfile->GetFileHash(),16);
						if (reqfile->IsPartFile())
							((CPartFile*)reqfile)->WritePartStatus(data);
						else{
							uint32 null = 0;
							data->Write(&null,3);
						}
						packet = new Packet(data);
						packet->opcode = OP_FILESTATUS;
						delete data;
						SendPacket(packet,true);
						break;
					}
					throw GetResString(IDS_ERR_WRONGPACKAGESIZE);
					break;
				}
				case OP_FILEREQANSWER:{
					client->ProcessFileInfo(packet,size);
					break;
				}
				case OP_FILESTATUS:{
					client->ProcessFileStatus(packet,size);
					break;
				}
				case OP_STARTUPLOADREQ:{
					theApp.uploadqueue->AddClientToQueue(client);
					break;
				}
				case OP_ACCEPTUPLOADREQ:{
					if (client->GetDownloadState() == DS_ONQUEUE ){
						client->SetDownloadState(DS_DOWNLOADING);
						client->SendBlockRequests();
					}
					break;
				}
				case OP_REQUESTPARTS:{
					CSafeMemFile* data = new CSafeMemFile((BYTE*)packet,size);
					uchar reqfilehash[16];
					data->Read(reqfilehash,16);
					Requested_Block_Struct* reqblock1 = new Requested_Block_Struct;
					Requested_Block_Struct* reqblock2 = new Requested_Block_Struct;
					Requested_Block_Struct* reqblock3 = new Requested_Block_Struct;
					data->Read(&reqblock1->StartOffset,4);
					data->Read(&reqblock2->StartOffset,4);
					data->Read(&reqblock3->StartOffset,4);
					data->Read(&reqblock1->EndOffset,4);
					data->Read(&reqblock2->EndOffset,4);
					data->Read(&reqblock3->EndOffset,4);
					memcpy(&reqblock1->FileID,reqfilehash,16);
					memcpy(&reqblock2->FileID,reqfilehash,16);
					memcpy(&reqblock3->FileID,reqfilehash,16);
					if (reqblock1->EndOffset-reqblock1->StartOffset == 0)			
						delete reqblock1;
					else
						client->AddReqBlock(reqblock1);

					if (reqblock2->EndOffset-reqblock2->StartOffset == 0)			
						delete reqblock2;
					else
						client->AddReqBlock(reqblock2);

					if (reqblock3->EndOffset-reqblock3->StartOffset == 0)			
						delete reqblock3;
					else
						client->AddReqBlock(reqblock3);
					delete data;
					break;
				}
				case OP_CANCELTRANSFER:{
					theApp.uploadqueue->RemoveFromUploadQueue(client);
					break;
				}
				case OP_HASHSETREQUEST:{
					if (size != 16)
						throw GetResString(IDS_ERR_WRONGHPACKAGESIZE);
					client->SendHashsetPacket(packet);
					break;
				}
				case OP_HASHSETANSWER:{
					client->ProcessHashSet(packet,size);
					break;
				}
				case OP_SENDINGPART:{
					client->ProcessBlockPacket(packet,size);
					break;
				}
				case OP_OUTOFPARTREQS:{
					break;
				}
				case OP_SETREQFILEID:{
					if (size == 16)
						memcpy(client->reqfileid,packet,16);
					break;
				}
				case OP_MESSAGE:{
					uint16 length;
					memcpy(&length,packet,2);
					if (length+2 != size)
						throw CString("invalid message packet");
					char* message = new char[length+1];
					memcpy(message,packet+2,length);
					message[length] = 0;
					theApp.emuledlg->chatwnd.chatselector.ProcessMessage(client,message);
					delete[] message;
					break;
				}
				case OP_ASKSHAREDFILES:
				{	// client wants to know what we have in share, let's see if we allow him to know that
					CPtrList list;
					if (theApp.glob_prefs->CanSeeShares() == 0 /*||					// everybody
						(theApp.glob_prefs->CanSeeShares() == 1 && IsFriend(client))*/ )	// friend
					{
						for (POSITION pos = theApp.sharedfiles->list.GetHeadPosition(); pos != NULL; )
						{
							CKnownFile* kf = (CKnownFile*)theApp.sharedfiles->list.GetNext(pos);
							list.AddTail((void*&)kf);
						}
					}
					// now create the memfile for the packet
					uint32 iTotalCount = list.GetCount();
					CSafeMemFile tempfile(100);
					tempfile.Write(&iTotalCount, 4);
					while (list.GetCount())
					{
						theApp.sharedfiles->CreateOfferedFilePacket((CKnownFile*)list.GetHead(), &tempfile);
						list.RemoveHead();
					}
					// create a packet and send it
					Packet* replypacket = new Packet(&tempfile);
					replypacket->opcode = OP_ASKSHAREDFILESANSWER;
					SendPacket(replypacket, true, false);
					break;
				}
				case OP_ASKSHAREDFILESANSWER:
				{	// client tells us what he has in share
					// todo display this one in a new, seperate tab, for now, this is sufficient
					theApp.searchlist->NewSearch(NULL, "Any");
					theApp.emuledlg->searchwnd.LocalSearchEnd(theApp.searchlist->ProcessSearchanswer(packet,size));
					break;
				}
				default:
					;//theApp.emuledlg->AddLogLine(false,"unknown opcode: %i %h",opcode,opcode);
				  
			}
		}
		catch(CFileException* error){
			error->Delete();	//mf
			throw GetResString(IDS_ERR_INVALIDPACKAGE);
		}
	}
	catch(CString error){
		if (client){
			client->SetDownloadState(DS_ERROR);	
			//TODO write this into a debugfile
#ifdef _DEBUG
			theApp.emuledlg->AddLogLine(false,GetResString(IDS_ERR_CLIENTERROR)
			,client->GetUserName(),client->GetFullIP(),error.GetBuffer());
#endif
		}
		else
			theApp.emuledlg->AddLogLine(false,GetResString(IDS_ERR_BADCLIENTACTION),error.GetBuffer());
			
		Disconnect();
		return false;
	}
	return true;
}

bool CClientReqSocket::ProcessExtPacket(char* packet, uint32 size, uint8 opcode){
	try{
		if (!client)
			throw GetResString(IDS_ERR_UNKNOWNCLIENTACTION);
		switch(opcode){
			case OP_EMULEINFO:{
				client->ProcessMuleInfoPacket(packet,size);
				client->SendMuleInfoPacket(true);
				break;
			}
			case OP_EMULEINFOANSWER:{
				client->ProcessMuleInfoPacket(packet,size);
				break;
			}
			case OP_COMPRESSEDPART:{
				client->ProcessBlockPacket(packet,size,true);
				break;
			}
			case OP_QUEUERANKING:{
				if (size != 12)
					throw GetResString(IDS_ERR_BADSIZE);
				uint16 newrank;
				memcpy(&newrank,packet+0,2);
				client->SetRemoteQueueRank(newrank);
			}
			/*case OP_VERIFYUPSREQ:{
				CUpDownClient* vuclient = theApp.clientlist->VUGetRandomClient();
				if (vuclient){
					Packet* answer = new Packet(OP_VERIFYUPSANSWER,6,OP_EMULEPROT);
					uint32 ip = vuclient->GetIP();
					memcpy(answer->pBuffer,ip,4);
					uint16 udpport = vuclient->GetUDPPort();
					memcpy(answer->pBuffer,ip,4);
					SendPacket(answer,true,true);
				}
			}
			case OP_VERIFYUPSANSWER:{
				if (size != 12)
					throw CString("invalid size (OP_VERIFYUPSANSWER)");
			}*/

			default:
				;//theApp.emuledlg->AddLogLine(false,"unknown opcode: %i %h",opcode,opcode);
			  
		}
	}
	catch(CString error){
		theApp.emuledlg->AddLogLine(false,GetResString(IDS_ERR_BADCLIENTACTION),error.GetBuffer());
		client->SetDownloadState(DS_ERROR);		
		Disconnect();
		return false;
	}
	return true;
}

void CClientReqSocket::OnInit(){
	uint8 tv = 1;
	//SetSockOpt(SO_DONTLINGER,&tv,sizeof(BOOL));
}

void CClientReqSocket::OnSend(int nErrorCode){
	ResetTimeOutTimer();
	CEMSocket::OnSend(nErrorCode);
}

void CClientReqSocket::OnError(int nErrorCode){

#ifdef _DEBUG
	if (client)
		theApp.emuledlg->AddLogLine(false,GetResString(IDS_ERR_BADCLIENT2)
		,client->GetUserName(),client->GetFullIP(),nErrorCode);
	else
		theApp.emuledlg->AddLogLine(false,GetResString(IDS_ERR_BADCLIENTACTION),nErrorCode);			
#endif
	Disconnect();
}

void CClientReqSocket::PacketReceived(Packet* packet){
	switch (packet->prot){
		case OP_EDONKEYPROT:
			ProcessPacket(packet->pBuffer,packet->size,packet->opcode);
			break;
		case OP_EXTENDEDPROT:
		case OP_EMULEPROT:
			ProcessExtPacket(packet->pBuffer,packet->size,packet->opcode);
			break;
		default:;
	}
}

void CClientReqSocket::OnReceive(int nErrorCode){
	ResetTimeOutTimer();
	CEMSocket::OnReceive(nErrorCode);
}

bool CClientReqSocket::Create(){
	BOOL result = CAsyncSocket::Create(0,SOCK_STREAM,FD_WRITE|FD_READ|FD_CLOSE);
	OnInit();
	return result;
}
// CListenSocket
// CListenSocket member functions
CListenSocket::CListenSocket(CPreferences* in_prefs){
	app_prefs = in_prefs;
	opensockets = 0;
}

CListenSocket::~CListenSocket(){
	Close();
	KillAllSockets();
}

bool CListenSocket::StartListening(){
	return (this->Create(app_prefs->GetPort(),SOCK_STREAM,FD_ACCEPT) && this->Listen());
}

void CListenSocket::OnAccept(int nErrorCode){
	//TODO: Block banned IPs here
	if (!nErrorCode && !( TooManySockets() && !theApp.serverconnect->IsConnecting() ) ){
		CClientReqSocket* newclient = new CClientReqSocket(app_prefs);
		if (!Accept(*newclient))
			newclient->Safe_Delete();
		else{
			newclient->AsyncSelect(FD_WRITE|FD_READ|FD_CLOSE);
			newclient->OnInit();
		}
	}
}

void CListenSocket::Process(){
	POSITION pos2;
	opensockets = 0;
	for(POSITION pos1 = socket_list.GetHeadPosition(); ( pos2 = pos1 ) != NULL; ){
		socket_list.GetNext(pos1);
		CClientReqSocket* cur_sock = socket_list.GetAt(pos2);
		if ( cur_sock->IsConnected() )
			opensockets++;

	   if (cur_sock->deletethis){
		   if (cur_sock->m_hSocket != INVALID_SOCKET){
				cur_sock->Close();
		   }
		   else{
			   cur_sock->Delete_Timed();;
		   }
	   }
	   else
		socket_list.GetAt( pos2 )->CheckTimeOut();
   }

}

void CListenSocket::AddSocket(CClientReqSocket* toadd){
	socket_list.AddTail(toadd);
}

void CListenSocket::RemoveSocket(CClientReqSocket* todel){
	POSITION pos2,pos1;
	for(pos1 = socket_list.GetHeadPosition(); ( pos2 = pos1 ) != NULL; ){
       socket_list.GetNext(pos1);
	   if ( socket_list.GetAt(pos2) == todel )
			socket_list.RemoveAt(pos2);
   }
}

void CListenSocket::KillAllSockets(){
	for (POSITION pos = socket_list.GetHeadPosition();pos != 0;pos = socket_list.GetHeadPosition()){
		CClientReqSocket* cur_socket = socket_list.GetAt(pos);
			if (cur_socket->client)
				delete cur_socket->client;
			else
				delete cur_socket;
	}
}

bool CListenSocket::TooManySockets(){
	if (GetOpenSockets() > app_prefs->GetMaxConnections()){
		return true;
	}
	else
		return false;
}

bool CListenSocket::IsValidSocket(CClientReqSocket* totest){
	return socket_list.Find(totest);
}

void CListenSocket::Debug_ClientDeleted(CUpDownClient* deleted){
	POSITION pos1, pos2;
	for (pos1 = socket_list.GetHeadPosition();( pos2 = pos1 ) != NULL;){
		socket_list.GetNext(pos1);
		CClientReqSocket* cur_sock = socket_list.GetAt(pos2);
		if (!AfxIsValidAddress(cur_sock, sizeof(CClientReqSocket)))
			AfxDebugBreak();
		if (cur_sock->client == deleted){
			AfxDebugBreak();
		}
	}
}
